"""P SQL Input Wizard

Help to create method, form and result page to insert data in a database.

This is a wizard to help create database input forms and SQL methods. 
The idea is to help the user to select the database tables and columns to 
which input data and the wizard automagically creates the insert SQL 
method and the HTML input form.

You'll probably have to edit the generated documents and method. The method 
specially if you are inputing data in multiple tables, since you'll want to 
have the same data in multiple dates (e.g. person_id).
Remember that it is easier to take out paramenters than to search your 
documentation/database for them, so don't be shy selecting items.

This software is released under GNU public license. See details in the URL:
http://www.gnu.org/copyleft/gpl.html

Author: Paulo Eduardo Neves <neves@inf.puc-rio.br>

Please send me bugs and feedback about how to improve it.
"""
import string
import Acquisition
from Globals import HTMLFile, Persistent, MessageDialog
from  Products.ZSQLMethods.SQL import *



def implementedSQLConnectionIDs(self):
    """Find SQL database connections in the current folder and above

    This function return a list of ids.
    Shameless stolen (with little modifications) from SQL.py
    """
    ids={}
    have_id=ids.has_key
    StringType=type('')

    while self is not None:
        if hasattr(self, 'objectValues'):
            for o in self.objectValues():
		try:
		    if (hasattr(o,'_isAnSQLConnection') 
			and o._isAnSQLConnection
			and hasattr(o,'id')):
			#let's see if this DB is suported
			if not hasattr(o, 'database_type'):
			    raise ImportError
			exec('import DBInfo.' + getattr(o, 'database_type'))
			
			id=o.id
			if type(id) is not StringType: id=id()
			if not have_id(id):
			    if hasattr(o,'title_and_id'): o=o.title_and_id()
			    else: o=id
			    ids[id]=id
		except ImportError:
		    #Database is not supported
		    pass 
        if hasattr(self, 'aq_parent'): self=self.aq_parent
        else: self=None

    ids=map(lambda item: (item[1], item[0]), ids.items())
    ids.sort()
    return ids

manage_startWizard = HTMLFile('selectDB', globals())

def _generateId(self, baseName):
    "Returns a not used attribute in the self object"
    id = baseName
    count = 1
    while hasattr(self, id):
	id = baseName + `count`
	count = count + 1
    return id

def debug(self):
    "debug"
    return ""

def manage_createWizard(self, connection_id, REQUEST):
    """Create the wizard for the desired connection_id
       and call it's default method """

    id = _generateId(self, "psqlwiz") #default name
    wiz = PSQLInputWizard(id, connection_id)
    setattr(self, id, wiz)

    REQUEST.RESPONSE.redirect("%s/%s" % (REQUEST['URL1'], id))

class Table:
    """Class to group table and columns atributtes

    Just to have a cleaner (dtml) code"""
    def __init__(self, name, columns, optionalColumns = []):
	self.tableName = name
	self.columns = columns
	self.optional = optionalColumns
	
    def __str__(self):
	return "Table Name:%s<BR>\nColumns:%s" % (self.tableName, `self.columns`)

class PSQLInputWizard(
    Acquisition.Implicit,
    Persistent,
    ):
    """A Wizard to create  forms and methods to insert data in databases

    The wizard uses a sequence of forms to ask the user for data,
    these data are stored as attributes in the wizard. When all the necessary 
    data are gathered it creates the input form, result page and method.
    """


    __ac_permissions__ = (
	('Use SQL Wizard', ('index_html', 'collectIds', 
			    'selectColumns', 'getObjectsInfo',
			    'createSQLInput', 'quit')),
	)
    meta_type = 'P SQL Input Wizard'



    dtmlDir = 'Products/PSQLInput/'

    def __init__(self, id, connection_id):
	self.id = id
	self.connection_id = connection_id
	self.tables = []
	self.selectedTables = []
	self.tableColumns = []
	self.selectedColumns = []
	self.sqlArguments = []



    def _infoFactory(self):
	connection = getattr(self, self.connection_id)
	exec('import DBInfo.' + connection.database_type)
	return eval('DBInfo.' + connection.database_type + '.Info(connection)')

    def index_html(self):
	"Start input wizard data gathering, show database tables to the user"

	self.info = self._infoFactory()
	self.tables = self.info.getTablesNames() 

	selectTables = HTMLFile( self.dtmlDir + 'selectTables')
	return selectTables(self, self)


    def queryColumns(self,  selectedTables = []):
	"Ask user to select data columns in selected tables"

	if selectedTables != []:
	    self.selectedTables = []
	    for i in selectedTables:
		self.selectedTables.append(self.tables[i])

	    self.tableColumns = []
	    for i, j in self.info.getColumnsNames(self.selectedTables).items():
		self.tableColumns.append(Table(i,j))

	selectColumns = HTMLFile(self.dtmlDir + 'selectColumns')
	return selectColumns(self,self)
	
    
    def _cleanColName(self, name):
	return string.join(map(string.capitalize, string.split(name, '_')))
    
    def _uniqueCol(self, table, column):
	#return table + string.capitalize(column)
	return table + '.' + column
 
    def _buildInputForm(self, resultPageId):
	#base %s parameters: result page id, form contents
	base = """<!--#var standard_html_header-->
<H2><!--#var title_or_id--> <!--#var document_title--></H2>
<P><FORM ACTION="%s" METHOD="POST">
<TABLE>
%s
</TABLE>
<input type="SUBMIT" name="submit" value="Insert">
</FORM>
<!--#var standard_html_footer-->"""
        inputBase = '\t<TR><TD ALIGN="RIGHT">%s</TD><TD><INPUT TYPE="TEXT" NAME="%s"></TD></TR>'
	form = []
	for i in self.selectedColumns:
	    form.append('<TR BGCOLOR="#CCCCCC"><TD COLSPAN=2>%s</TD></TR>' 
			% i.tableName)
	    for j in i.columns:
		if j in i.optional:
		    form.append(inputBase 
				% (self._cleanColName(j), 
				   self._uniqueCol(i.tableName, j)))
		else:
		    form.append(inputBase 
				% ("<B>" + self._cleanColName(j) + "</B>", 
				   self._uniqueCol(i.tableName, j)))


	return base % (resultPageId, string.join(form, '\n'))

    def _buildResultPage(self, sqlMethodId):
	base = """<!--#var standard_html_header-->
<H2><!--#var title_or_id--> <!--#var document_title--></H2>
<P><!--#call "%s(REQUEST)"-->
Your data was inserted in the Database!
<!--#var standard_html_footer-->""" # How do I catch SQL errors?
        return base % sqlMethodId

    def debug(self):
	"sss"
	x = ''
	for i in table.columns: x= x+ ' ' + i
	x = x+ '\n************\n'
	for i in table.optional: x= x+ ' ' + i
	return x

    def _buildSQLValues(self, table, nonOptCol):
	template = '<!--#sqlvar %s type=%s %s-->'
	args = []
	for i in table.columns:
	    tp = self.info.type(table.tableName, i)
	    opt = self.info.nullable(table.tableName, i)
	    uniqueCol = self._uniqueCol(table.tableName, i)
	    
	    oneArg = template % (uniqueCol, tp, opt) + ','
	    if i in table.optional:
		args.append('<!--#if %s-->\n\t\t\t%s\n\t\t<!--#/if-->'
			    % (uniqueCol, oneArg))
	    else:
		args.append(oneArg)

	#there's just missing the non optional (without the comma)
	tp = self.info.type(table.tableName, nonOptCol)
	opt = self.info.nullable(table.tableName, nonOptCol)
	uniqueCol = self._uniqueCol(table.tableName, nonOptCol)
	args.append(template % (uniqueCol, tp, opt))

	return string.join(args, '\n\t\t')

    def _buildSQLInto(self, table, nonOptCol):
	tempCols = []
	for i in table.columns:
	    tempCols.append(i + ',')

	newCol = []
	for i in range(len(table.columns)):
	    if  table.columns[i] in table.optional:
		newCol.append('<!--#if %s -->\n\t\t%s\n\t<!--#/if-->' % 
			      (self._uniqueCol(table.tableName,
					       table.columns[i]), 
			       tempCols[i]))
	    else:
		newCol.append(tempCols[i])

	newCol.append(nonOptCol)
	return string.join(newCol,'\n\t')
	
    def _buildSQLInputCommand(self):
	baseCommand = "INSERT INTO %s \n\t(%s)\n\tVALUES (%s)"
	command = []
	inputArgs = []

	for i in self.selectedColumns:
	    #building sql method arguments
	    for j in i.columns:
		tempArg = self._uniqueCol(i.tableName, j)
		if j in i.optional:
		    tempArg = tempArg + '=""'
		inputArgs.append(tempArg)

	    #Let's find a column that isn't optional, 
	    #I've already verified in getObjectsInfo that it exists
	    for col in i.columns:
		if col not in i.optional:
		    nonOptCol = col
		    i.columns.remove(col)
		    break

	    #building sql method
	    command.append(baseCommand % (i.tableName, 
					  self._buildSQLInto(i, nonOptCol),
					  self._buildSQLValues(i, nonOptCol)))

	return string.join(inputArgs), string.join(command,'\n<!--#var sql_delimiter -->\n')


    def getObjectsInfo(self, REQUEST):
	"Store data and Collect id's and title of objects to be created"
	self.selectedColumns = []
	for i in self.selectedTables:
	    if REQUEST.has_key(i):
		optional = []
		if REQUEST.has_key('opt.' + i):
		    optional = REQUEST['opt.' + i]

		oneColNonOptional = None
		for j in REQUEST[i]:
		    if j not in optional:
			oneColNonOptional = 1
			break
		if oneColNonOptional is None:
		    return MessageDialog(title="Error",
					 message = "<h1>Error</h1>There must be at least one selected non-optional column",
					 action = REQUEST['HTTP_REFERER']
					 )
		
		self.selectedColumns.append(Table(i, REQUEST[i], 
						  optional))

	if self.selectedColumns == []:
	    return MessageDialog(title="Error",
				 message = "<h1>Error</h1>You Must select at least one column of one table",
				 action = REQUEST['HTTP_REFERER']
				 )

	allTables = string.join(map(string.capitalize, self.selectedTables),'')
	defaultTitles = string.join(map(string.capitalize, 
					self.selectedTables))
	formDefaultId = _generateId(self.aq_parent, allTables + "InsertForm")
	resultDefaultId = _generateId(self.aq_parent, 
				      allTables + "InsertResult")
	sqlMethodDefaultId = _generateId(self.aq_parent, 
					 allTables + "InsertMethod")

	collectIds = HTMLFile(self.dtmlDir + 'collectIds')
	return  collectIds(self, 
			   formId=formDefaultId, 
			   resultId=resultDefaultId,
			   sqlId=sqlMethodDefaultId,
			   titles=defaultTitles)
    
    def createSQLInput(self, RESPONSE, URL2, formId, resultId, sqlId, 
		       formTitle='', resultTitle = '', sqlTitle = ''):
	"""Method to add the Z SQL Method and a Form to access it, 
	also delete Wizard from parent folder """

	try:
	    self.aq_parent.manage_addDTMLDocument(formId, formTitle, 
						  self._buildInputForm(resultId) ) 
	except AttributeError:
	    self.aq_parent.manage_addDocument(formId, formTitle, 
						  self._buildInputForm(resultId) ) 

	try:
	    self.aq_parent.manage_addDTMLDocument(resultId, resultTitle, 
						  self._buildResultPage(sqlId) ) 
	except AttributeError:
	    self.aq_parent.manage_addDocument(resultId, resultTitle, 
						  self._buildResultPage(sqlId) ) 
	    

	args, command = self._buildSQLInputCommand()
	#self.aq_parent.manage_addZSQLMethod(sqlId, sqlTitle, 
	#				    self.connection_id,
	#				    args, command)

	self.aq_parent._setObject(sqlId, SQL(sqlId, sqlTitle,
		self.connection_id, args, command))

	self.quit(RESPONSE, URL2)

    def quit(self, RESPONSE, URL2):
	"Cleans temporary objects used by P SQL Wizard"
	
	delattr(self.aq_parent, self.id) #deleting wizard, if you leave in the middle there will be trash in your folder
	
	
	RESPONSE.redirect(URL2 + '/manage_main') #should I show some message?


    

